// Copyright 2014 Canonical Ltd.
// Licensed under the AGPLv3, see LICENCE file for details.

package debugger_test

import (
	"strconv"
	"time"

	gc "launchpad.net/gocheck"

	"launchpad.net/juju-core/juju/testing"
	"launchpad.net/juju-core/names"
	"launchpad.net/juju-core/state/api/params"
	"launchpad.net/juju-core/state/apiserver/common"
	"launchpad.net/juju-core/state/apiserver/debugger"
	apiservertesting "launchpad.net/juju-core/state/apiserver/testing"
	statetesting "launchpad.net/juju-core/state/testing"
	coretesting "launchpad.net/juju-core/testing"
)

type debuggerSuite struct {
	testing.JujuConnSuite

	authorizer apiservertesting.FakeAuthorizer
	resources  *common.Resources
	debugger   *debugger.DebuggerAPI
}

var _ = gc.Suite(&debuggerSuite{})

func (s *debuggerSuite) SetUpTest(c *gc.C) {
	s.JujuConnSuite.SetUpTest(c)

	// Create a FakeAuthorizer so we can check permissions.
	s.authorizer = apiservertesting.FakeAuthorizer{
		Tag:            names.EnvironTag("fake"),
		LoggedIn:       true,
		EnvironManager: true,
	}

	// Create the resource registry separately to track invocations to
	// Register.
	s.resources = common.NewResources()

	// Create a debugger API for the environment.
	debugger, err := debugger.NewDebuggerAPI(
		s.State,
		s.resources,
		s.authorizer,
	)
	c.Assert(err, gc.IsNil)
	s.debugger = debugger
}

// watchDebugLogTests is a subset of logTailerTests to verify that lines and tag filtering
// are working.
var watchDebugLogTests = []struct {
	description       string
	lines             int
	tag               string
	initialCollected  []string
	appendedCollected []string
}{{
	description:      "zero initial lines, no filtering of a tag",
	lines:            0,
	tag:              "",
	initialCollected: []string{},
	appendedCollected: []string{
		"unit-wordpress-0: 2013-12-19 12:30:11 INFO worker.uniter.jujuc server.go:1 alpha bravo charlie\n",
		"unit-wordpress-0: 2013-12-19 12:30:12 DEBUG worker.uniter.jujuc server.go:2 delta echo foxtrott\n",
		"machine-0: 2013-12-19 12:30:30 DEBUG juju main.go:3 developer wants to know something else\n",
		"unit-mysql-0: 2013-12-19 12:33:44 DEBUG juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-mysql-0: 2013-12-19 12:33:44 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-mysql-0\n",
		"unit-mysql-0: 2013-12-19 12:33:45 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-wordpress-0\n",
		"unit-wordpress-0: 2013-12-19 12:34:00 INFO juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-wordpress-0: 2013-12-19 12:34:01 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"unit-wordpress-0: 2013-12-19 12:34:02 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:34:56 INFO juju main.go:1 something happens\n",
		"unit-mysql-0: 2013-12-19 12:40:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"unit-wordpress-0: 2013-12-19 12:41:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"machine-0: 2013-12-19 13:00:00 INFO juju.api.logger logtailer.go:99 end of test data\n",
	},
}, {
	description: "five initial lines, no filtering of a tag",
	lines:       5,
	tag:         "",
	initialCollected: []string{
		"machine-0: 2013-12-19 12:01:00 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:01 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:05 DEBUG juju main.go:2 developer wants to know something\n",
		"unit-mysql-0: 2013-12-19 12:05:00 ERROR worker.uniter.jujuc server.go:99 ouch\n",
		"machine-0: 2013-12-19 12:05:01 ERROR juju main.go:2 \"ouch\" happened at \"unit-mysql-0\"\n",
	},
	appendedCollected: []string{
		"unit-wordpress-0: 2013-12-19 12:30:11 INFO worker.uniter.jujuc server.go:1 alpha bravo charlie\n",
		"unit-wordpress-0: 2013-12-19 12:30:12 DEBUG worker.uniter.jujuc server.go:2 delta echo foxtrott\n",
		"machine-0: 2013-12-19 12:30:30 DEBUG juju main.go:3 developer wants to know something else\n",
		"unit-mysql-0: 2013-12-19 12:33:44 DEBUG juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-mysql-0: 2013-12-19 12:33:44 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-mysql-0\n",
		"unit-mysql-0: 2013-12-19 12:33:45 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-wordpress-0\n",
		"unit-wordpress-0: 2013-12-19 12:34:00 INFO juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-wordpress-0: 2013-12-19 12:34:01 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"unit-wordpress-0: 2013-12-19 12:34:02 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:34:56 INFO juju main.go:1 something happens\n",
		"unit-mysql-0: 2013-12-19 12:40:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"unit-wordpress-0: 2013-12-19 12:41:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"machine-0: 2013-12-19 13:00:00 INFO juju.api.logger logtailer.go:99 end of test data\n",
	},
}, {
	description: "five initial lines, filtering of machine-0, five matching initial lines expected",
	lines:       5,
	tag:         "machine-0",
	initialCollected: []string{
		"machine-0: 2013-12-19 12:00:01 DEBUG juju.api.logger logtailer.go:3 jumps over the lazy dog\n",
		"machine-0: 2013-12-19 12:01:00 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:01 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:05 DEBUG juju main.go:2 developer wants to know something\n",
		"machine-0: 2013-12-19 12:05:01 ERROR juju main.go:2 \"ouch\" happened at \"unit-mysql-0\"\n",
	},
	appendedCollected: []string{
		"machine-0: 2013-12-19 12:30:30 DEBUG juju main.go:3 developer wants to know something else\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-mysql-0\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-wordpress-0\n",
		"machine-0: 2013-12-19 12:34:56 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 13:00:00 INFO juju.api.logger logtailer.go:99 end of test data\n",
	},
}}

func (s *debuggerSuite) TestWatchDebugLog(c *gc.C) {
	for i, test := range watchDebugLogTests {
		c.Logf("test #%d: %s", i, test.description)
		_, sigc, restorer := startLogDataWriter(c)
		c.Assert(s.resources.Count(), gc.Equals, i)

		if test.tag == "" {
			// No tag set, so tetrieve the environ tag for test.
			environ, err := s.State.Environment()
			c.Assert(err, gc.IsNil)
			test.tag = environ.Tag()
		}

		args := params.EntityLogRequests{Entities: []params.EntityLogRequest{
			{Tag: test.tag, Lines: test.lines},
		}}

		result := s.debugger.WatchDebugLog(args)
		id := strconv.Itoa(i + 1)
		c.Assert(result, gc.DeepEquals, params.StringsWatchResults{
			Results: []params.StringsWatchResult{
				{Changes: test.initialCollected, StringsWatcherId: id},
			},
		})

		// Verify the resource was registered and stop when done.
		c.Assert(s.resources.Count(), gc.Equals, i+1)
		c.Assert(result.Results[0].StringsWatcherId, gc.Equals, id)
		resource := s.resources.Get(id)
		defer statetesting.AssertStop(c, resource)

		// Check that the Watch has consumed the initial event ("returned" in
		// the Watch call).
		wc := statetesting.NewStringsWatcherC(c, s.State, resource.(statetesting.StringsWatcher))
		wc.AssertNoChange()

		sigc <- struct{}{}

		assertWatched(c, resource.(statetesting.StringsWatcher), test.appendedCollected)

		restorer.Restore()
	}
}

// assertWatched collects the changes returned by the debug log watch and compares
// them to the passed test data.
func assertWatched(c *gc.C, watcher statetesting.StringsWatcher, compare []string) {
	timeout := time.After(coretesting.LongWait)
	lines := []string{}
	for {
		select {
		case changes, ok := <-watcher.Changes():
			if ok {
				if compare == nil {
					// Didn't expected data, so fail.
					c.Fatalf("collected unexpected data")
				}
				lines = append(lines, changes...)
				if len(lines) == len(compare) {
					// All lines received.
					c.Assert(lines, gc.DeepEquals, compare)
					return
				}
			} else {
				// changesc closed after stopping or error.
				c.Assert(lines, gc.DeepEquals, compare[:len(lines)])
				return
			}
		case <-timeout:
			if compare != nil {
				// Expected data, so timeout means failing.
				c.Fatalf("timeout during debug log collection")
			}
			return
		}
	}
}
