// Copyright 2014 Canonical Ltd.
// Licensed under the AGPLv3, see LICENCE file for details.

package debugger

import (
	"bytes"
	"fmt"
	"os"
	"time"

	"launchpad.net/tomb"

	"launchpad.net/juju-core/utils/tailer"
)

// logTailer reads the combined all machines log writes the
// filtered data into a channel of string slices.
type logTailer struct {
	logFile *os.File
	writer  *channelWriter
	tailer  *tailer.Tailer
}

// startLogTailer opens the passed log file and tails the lines
// with the passed prefix.
func startLogTailer(filename string, lines, maxLines int, filter tailer.FilterFunc) (*logTailer, error) {
	logFile, err := os.Open(filename)
	if err != nil {
		return nil, err
	}
	if maxLines > 0 && maxLines < lines {
		// Maximum number of lines has to be at least same size as
		// number of initial lines, otherwise those would be truncated.
		maxLines = lines
	}
	lt := &logTailer{
		logFile: logFile,
		writer:  startChannelWriter(maxLines),
	}
	lt.tailer = tailer.NewTailer(lt.logFile, lt.writer, lines, filter)
	return lt, nil
}

// Changes returns the changes to the tailed log. These are the filtered
// appended lines.
func (lt *logTailer) Changes() <-chan []string {
	return lt.writer.Lines()
}

// Stop terminates the log tailer by closing in- and output
// after stopping the tailer.
func (lt *logTailer) Stop() error {
	defer func() {
		err := lt.writer.Close()
		if err != nil {
			logger.Errorf("cannot close channel writer: %v", err)
		}
	}()
	defer func() {
		err := lt.logFile.Close()
		if err != nil {
			logger.Errorf("cannot close log file: %v", err)
		}
	}()
	return lt.tailer.Stop()
}

// Err implements the Errer interface.
func (lt *logTailer) Err() error {
	return lt.tailer.Err()
}

// channelWriter implements an io.Writer which passes the written
// data into a channel of string slices.
type channelWriter struct {
	tomb     tomb.Tomb
	in       chan []byte
	out      chan []string
	maxLines int
}

// startChannelWriter creates a Writer converting its data into slices
// of strings, each terminated by newlines. maxLines controls the maximum
// number of lines kept between retrieving to avoid a memory leak if
// the data isn't retrieved. A number below 1 means all lines.
func startChannelWriter(maxLines int) *channelWriter {
	cw := &channelWriter{
		in:       make(chan []byte),
		out:      make(chan []string),
		maxLines: maxLines,
	}
	go func() {
		defer cw.tomb.Done()
		defer close(cw.out)
		cw.tomb.Kill(cw.loop())
	}()
	return cw
}

// Write implements the Writer interface.
func (cw *channelWriter) Write(p []byte) (int, error) {
	select {
	case cw.in <- p:
	case <-cw.tomb.Dying():
		return 0, fmt.Errorf("channel writer is closed")
	}
	return len(p), nil
}

// Lines returns a channel delivering the lines written
// to the writer.
func (cw *channelWriter) Lines() <-chan []string {
	return cw.out
}

// Close closes the writer.
func (cw *channelWriter) Close() error {
	cw.tomb.Kill(nil)
	return cw.tomb.Wait()
}

// loop is the backend of the channel writer.
func (cw *channelWriter) loop() error {
	lines := []string{}
	buffer := []byte(nil)
	timer := time.NewTimer(10 * time.Millisecond)
	defer timer.Stop()
	trySend := func() {
		if lines != nil {
			select {
			case cw.out <- lines:
				lines = nil
			default:
			}
		}
	}
	for {
		select {
		case <-cw.tomb.Dying():
			return tomb.ErrDying
		case block := <-cw.in:
			// Extract lines and send them.
			lines, buffer = cw.split(lines, buffer, block)
			trySend()
		case <-timer.C:
			// Send collected lines and reset timer.
			trySend()
			timer.Reset(10 * time.Millisecond)
		}
	}
}

// split splits the combination of the former rest buffer and a
// new block into lines and returns them with possible older not yet
// delivered lines. The rest of the block which cannot be splitted
// is returned as new buffer.
func (cw *channelWriter) split(lines []string, buffer, block []byte) ([]string, []byte) {
	if len(block) == 0 {
		return lines, buffer
	}
	work := append(buffer, block...)
	start := 0
	rest := []byte(nil)
	for {
		i := bytes.IndexByte(work[start:], '\n')
		if i < 0 {
			// Block terminates without a newline.
			rest = work[start:]
			break
		}
		lines = append(lines, string(work[start:start+i+1]))
		start += i + 1
		if start >= len(work) {
			// Block terminates with a newline.
			break
		}
	}
	// Check length of lines.
	if cw.maxLines > 0 && len(lines) > cw.maxLines {
		lines = lines[len(lines)-cw.maxLines:]
	}
	return lines, rest
}
