// Copyright 2014 Canonical Ltd.
// Licensed under the AGPLv3, see LICENCE file for details.

package debugger_test

import (
	"bytes"
	"time"

	gc "launchpad.net/gocheck"

	"launchpad.net/juju-core/state/apiserver/debugger"
	"launchpad.net/juju-core/testing"
)

type logTailerSuite struct{}

var _ = gc.Suite(&logTailerSuite{})

var channelWriterTests = []struct {
	description string
	data        []byte
	writings    []int
	readings    [][]string
}{{
	description: "terminated lines, one by one",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog\n"),
	writings:    []int{20, 35, 44},
	readings: [][]string{
		{"the quick brown fox\n"},
		{"jumps over the\n"},
		{"lazy dog\n"},
	},
}, {
	description: "terminated lines, in groups",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog\n"),
	writings:    []int{35, 44},
	readings: [][]string{
		{"the quick brown fox\n", "jumps over the\n"},
		{"lazy dog\n"},
	},
}, {
	description: "one large block",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog\n"),
	writings:    []int{44},
	readings: [][]string{
		{"the quick brown fox\n", "jumps over the\n", "lazy dog\n"},
	},
}, {
	description: "two blocks containing an unterminated line",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog\n"),
	writings:    []int{10, 44},
	readings: [][]string{
		nil,
		{"the quick brown fox\n", "jumps over the\n", "lazy dog\n"},
	},
}, {
	description: "three blocks containing unterminated lines",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog\n"),
	writings:    []int{10, 30, 44},
	readings: [][]string{
		nil,
		{"the quick brown fox\n"},
		{"jumps over the\n", "lazy dog\n"},
	},
}, {
	description: "unterminated last line, one by one",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog"),
	writings:    []int{20, 35, 43},
	readings: [][]string{
		{"the quick brown fox\n"},
		{"jumps over the\n"},
		nil,
	},
}, {
	description: "empty lines or lines containing only spaces between the lines",
	data:        []byte("the quick brown fox\n\njumps over the\n\n       \nlazy dog\n"),
	writings:    []int{10, 35, 54},
	readings: [][]string{
		nil,
		{"the quick brown fox\n", "\n"},
		{"jumps over the\n", "\n", "       \n", "lazy dog\n"},
	},
}, {
	description: "writing line by line, asynchronous reading",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog\n"),
	writings:    []int{20, 35, 44},
	readings: [][]string{
		nil,
		nil,
		nil,
		{"the quick brown fox\n", "jumps over the\n", "lazy dog\n"},
	},
}, {
	description: "three blocks containing unterminated lines, asynchronous reading",
	data:        []byte("the quick brown fox\njumps over the\nlazy dog\n"),
	writings:    []int{10, 30, 44},
	readings: [][]string{
		nil,
		nil,
		nil,
		{"the quick brown fox\n", "jumps over the\n", "lazy dog\n"},
	},
}, {
	description: "one block, containing more lines than the maximum of 5, so only the last ones",
	data:        []byte("the\nquick\nbrown\nfox\njumps\nover\nthe\nlazy\ndog\n"),
	writings:    []int{44},
	readings: [][]string{
		{"jumps\n", "over\n", "the\n", "lazy\n", "dog\n"},
	},
}}

func (s *logTailerSuite) TestChannelWriter(c *gc.C) {
	for i, test := range channelWriterTests {
		c.Logf("test #%d: %s", i, test.description)
		cw := debugger.StartChannelWriter(5)
		// Run writings and readings.
		start := 0
		for j, end := range test.writings {
			writing := test.data[start:end]
			reading := test.readings[j]
			cw.Write(writing)
			if reading != nil {
				lines := <-cw.Lines()
				c.Assert(lines, gc.DeepEquals, reading)
			}
			start = end
		}
		// One more reading then writings means that all changes
		// are read in one block.
		if len(test.readings) > len(test.writings) {
			reading := test.readings[len(test.readings)-1]
			lines := <-cw.Lines()
			c.Assert(lines, gc.DeepEquals, reading)
		}
		c.Assert(cw.Close(), gc.IsNil)
	}
}

func filterNothing(line []byte) bool {
	return true
}

func makeTagFilter(tag string) func([]byte) bool {
	prefix := []byte(tag + ":")
	return func(line []byte) bool {
		return bytes.HasPrefix(line, prefix)
	}
}

var logTailerTests = []struct {
	description       string
	lines             int
	maxLines          int
	filter            func(line []byte) bool
	initialCollected  []string
	appendedCollected []string
}{{
	description:      "zero initial lines, no filtering of a tag",
	lines:            0,
	maxLines:         0,
	filter:           filterNothing,
	initialCollected: []string{},
	appendedCollected: []string{
		"unit-wordpress-0: 2013-12-19 12:30:11 INFO worker.uniter.jujuc server.go:1 alpha bravo charlie\n",
		"unit-wordpress-0: 2013-12-19 12:30:12 DEBUG worker.uniter.jujuc server.go:2 delta echo foxtrott\n",
		"machine-0: 2013-12-19 12:30:30 DEBUG juju main.go:3 developer wants to know something else\n",
		"unit-mysql-0: 2013-12-19 12:33:44 DEBUG juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-mysql-0: 2013-12-19 12:33:44 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-mysql-0\n",
		"unit-mysql-0: 2013-12-19 12:33:45 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-wordpress-0\n",
		"unit-wordpress-0: 2013-12-19 12:34:00 INFO juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-wordpress-0: 2013-12-19 12:34:01 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"unit-wordpress-0: 2013-12-19 12:34:02 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:34:56 INFO juju main.go:1 something happens\n",
		"unit-mysql-0: 2013-12-19 12:40:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"unit-wordpress-0: 2013-12-19 12:41:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"machine-0: 2013-12-19 13:00:00 INFO juju.api.logger logtailer.go:99 end of test data\n",
	},
}, {
	description: "five initial lines, no filtering of a tag",
	lines:       5,
	maxLines:    0,
	filter:      filterNothing,
	initialCollected: []string{
		"machine-0: 2013-12-19 12:01:00 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:01 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:05 DEBUG juju main.go:2 developer wants to know something\n",
		"unit-mysql-0: 2013-12-19 12:05:00 ERROR worker.uniter.jujuc server.go:99 ouch\n",
		"machine-0: 2013-12-19 12:05:01 ERROR juju main.go:2 \"ouch\" happened at \"unit-mysql-0\"\n",
	},
	appendedCollected: []string{
		"unit-wordpress-0: 2013-12-19 12:30:11 INFO worker.uniter.jujuc server.go:1 alpha bravo charlie\n",
		"unit-wordpress-0: 2013-12-19 12:30:12 DEBUG worker.uniter.jujuc server.go:2 delta echo foxtrott\n",
		"machine-0: 2013-12-19 12:30:30 DEBUG juju main.go:3 developer wants to know something else\n",
		"unit-mysql-0: 2013-12-19 12:33:44 DEBUG juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-mysql-0: 2013-12-19 12:33:44 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-mysql-0\n",
		"unit-mysql-0: 2013-12-19 12:33:45 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-wordpress-0\n",
		"unit-wordpress-0: 2013-12-19 12:34:00 INFO juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-wordpress-0: 2013-12-19 12:34:01 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"unit-wordpress-0: 2013-12-19 12:34:02 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:34:56 INFO juju main.go:1 something happens\n",
		"unit-mysql-0: 2013-12-19 12:40:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"unit-wordpress-0: 2013-12-19 12:41:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"machine-0: 2013-12-19 13:00:00 INFO juju.api.logger logtailer.go:99 end of test data\n",
	},
}, {
	description: "five initial lines, filtering of machine-0, five matching initial lines expected",
	lines:       5,
	maxLines:    0,
	filter:      makeTagFilter("machine-0"),
	initialCollected: []string{
		"machine-0: 2013-12-19 12:00:01 DEBUG juju.api.logger logtailer.go:3 jumps over the lazy dog\n",
		"machine-0: 2013-12-19 12:01:00 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:01 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:05 DEBUG juju main.go:2 developer wants to know something\n",
		"machine-0: 2013-12-19 12:05:01 ERROR juju main.go:2 \"ouch\" happened at \"unit-mysql-0\"\n",
	},
	appendedCollected: []string{
		"machine-0: 2013-12-19 12:30:30 DEBUG juju main.go:3 developer wants to know something else\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-mysql-0\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-wordpress-0\n",
		"machine-0: 2013-12-19 12:34:56 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 13:00:00 INFO juju.api.logger logtailer.go:99 end of test data\n",
	},
}, {
	description: "five initial lines, filtering of unit-mysql-0, three matching initial lines expected",
	lines:       5,
	maxLines:    0,
	filter:      makeTagFilter("unit-mysql-0"),
	initialCollected: []string{
		"unit-mysql-0: 2013-12-19 12:00:10 INFO worker.uniter.jujuc server.go:1 alpha bravo charlie\n",
		"unit-mysql-0: 2013-12-19 12:00:10 DEBUG worker.uniter.jujuc server.go:2 delta echo foxtrott\n",
		"unit-mysql-0: 2013-12-19 12:05:00 ERROR worker.uniter.jujuc server.go:99 ouch\n",
	},
	appendedCollected: []string{
		"unit-mysql-0: 2013-12-19 12:33:44 DEBUG juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-mysql-0: 2013-12-19 12:33:44 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"unit-mysql-0: 2013-12-19 12:33:45 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"unit-mysql-0: 2013-12-19 12:40:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
	},
}, {
	description:      "five initial lines, filtering of unit-wordpress-0, no matching initial lines expected",
	lines:            5,
	maxLines:         0,
	filter:           makeTagFilter("unit-wordpress-0"),
	initialCollected: []string{},
	appendedCollected: []string{
		"unit-wordpress-0: 2013-12-19 12:30:11 INFO worker.uniter.jujuc server.go:1 alpha bravo charlie\n",
		"unit-wordpress-0: 2013-12-19 12:30:12 DEBUG worker.uniter.jujuc server.go:2 delta echo foxtrott\n",
		"unit-wordpress-0: 2013-12-19 12:34:00 INFO juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-wordpress-0: 2013-12-19 12:34:01 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"unit-wordpress-0: 2013-12-19 12:34:02 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"unit-wordpress-0: 2013-12-19 12:41:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
	},
}, {
	description:       "five initial lines, filtering of machine-1, nothing expected",
	lines:             5,
	maxLines:          0,
	filter:            makeTagFilter("machine-1"),
	initialCollected:  []string{},
	appendedCollected: nil,
}, {
	description: "five initial lines, maximum number is lower, no filtering of a tag",
	lines:       5,
	maxLines:    2,
	filter:      filterNothing,
	initialCollected: []string{
		"machine-0: 2013-12-19 12:01:00 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:01 INFO juju main.go:1 something happens\n",
		"machine-0: 2013-12-19 12:01:05 DEBUG juju main.go:2 developer wants to know something\n",
		"unit-mysql-0: 2013-12-19 12:05:00 ERROR worker.uniter.jujuc server.go:99 ouch\n",
		"machine-0: 2013-12-19 12:05:01 ERROR juju main.go:2 \"ouch\" happened at \"unit-mysql-0\"\n",
	},
	appendedCollected: []string{
		"unit-wordpress-0: 2013-12-19 12:30:11 INFO worker.uniter.jujuc server.go:1 alpha bravo charlie\n",
		"unit-wordpress-0: 2013-12-19 12:30:12 DEBUG worker.uniter.jujuc server.go:2 delta echo foxtrott\n",
		"machine-0: 2013-12-19 12:30:30 DEBUG juju main.go:3 developer wants to know something else\n",
		"unit-mysql-0: 2013-12-19 12:33:44 DEBUG juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-mysql-0: 2013-12-19 12:33:44 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-mysql-0\n",
		"unit-mysql-0: 2013-12-19 12:33:45 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:33:45 DEBUG juju.state.apiserver apiserver.go:102 <- [1] unit-wordpress-0\n",
		"unit-wordpress-0: 2013-12-19 12:34:00 INFO juju.worker.uniter modes.go:1 HOOK foo\n",
		"unit-wordpress-0: 2013-12-19 12:34:01 INFO juju.worker.uniter modes.go:2 HOOK bar\n",
		"unit-wordpress-0: 2013-12-19 12:34:02 INFO juju.worker.uniter modes.go:3 HOOK yadda\n",
		"machine-0: 2013-12-19 12:34:56 INFO juju main.go:1 something happens\n",
		"unit-mysql-0: 2013-12-19 12:40:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"unit-wordpress-0: 2013-12-19 12:41:00 INFO juju.worker.uniter context.go:1 HOOK baz\n",
		"machine-0: 2013-12-19 13:00:00 INFO juju.api.logger logtailer.go:99 end of test data\n",
	},
}}

func (s *logTailerSuite) TestLogTailer(c *gc.C) {
	for i, test := range logTailerTests {
		c.Logf("test #%d: %s", i, test.description)
		logFileName, sigc, restorer := startLogDataWriter(c)
		logTailer, err := debugger.StartLogTailer(logFileName, test.lines, 0, test.filter)
		c.Assert(err, gc.IsNil)

		assertCollected(c, logTailer.Changes(), test.initialCollected)

		sigc <- struct{}{}

		assertCollected(c, logTailer.Changes(), test.appendedCollected)

		c.Assert(logTailer.Stop(), gc.IsNil)
		restorer.Restore()
	}
}

// assertCollected collects the changes returned by the LogTailer and compares
// them to the passed test data.
func assertCollected(c *gc.C, changesc <-chan []string, compare []string) {
	timeout := time.After(testing.LongWait)
	lines := []string{}
	for {
		select {
		case changes, ok := <-changesc:
			if ok {
				if compare == nil {
					// Didn't expected data, so fail.
					c.Fatalf("collected unexpected data")
				}
				lines = append(lines, changes...)
				if len(lines) == len(compare) {
					// All lines received.
					c.Assert(lines, gc.DeepEquals, compare)
					return
				}
			} else {
				// changesc closed after stopping or error.
				c.Assert(lines, gc.DeepEquals, compare[:len(lines)])
				return
			}
		case <-timeout:
			if compare != nil {
				// Expected data, so timeout means failing.
				c.Fatalf("timeout during log tailer collection")
			}
			return
		}
	}
}
